import numpy as np
import pandas as pd

import pyspark.sql.functions as F
import pyspark.sql.types as T

def search(indexed, search_for):
    """
    Given a 1-d array (indexed) and a search 1-d array (search_for),
      returns a 1-d array shaped like search_for where the values in search_for
      appear in indexed; non-matching elements are assigned to -1.
    """
    cross = np.equal(indexed, search_for[:,np.newaxis])
    results = np.argwhere(cross)
    indices = results[:, 0]
    values = results[:, 1]
    output = np.ones_like(search_for) * -1
    output[indices] = values
    return output

def dcg(best, output):
    """Computes discounted cumulative gain
    given ground-truth "best" ranking and output, 
    the re-ranked output of the model.
    """
    positions = np.arange(output.shape[0])
    rels = (positions.max() - positions) + 1
    rels = np.concatenate((rels, np.array([0])))
    
    idx = search(best, output)
    relevances = rels[idx]
    
    i = np.arange(1, len(best)+1)
    
    brp = (2.0 ** relevances)
    
    score = ((brp - 1)/np.log2(i + 1)).sum()
    
    return score

def ndcg(best, output):
    """Computes normalized discounted cumulative gain
    given ground-truth "best" ranking and output, 
    the re-ranked output of the model.
    """
    best, output = np.array(best), np.array(output)
    top_score = dcg(best, best)
    dcg_ = dcg(best, output)
    
    return float(dcg_/top_score)

def ranking_ndcg_at_k(best, other, k=None):
    """
    Computes ndcg at k using raw position
    as relevance score.
    """
    if k is None:
        k = len(other)
    other = other[:k]
    diff = len(other) - len(best)
    if diff == 0:
        pass
    elif diff < 0:
        other += ['_ ' for x in range(np.abs(diff))]
    elif diff > 0:
        best += [' _' for x in range(diff)]
    other = [hash(x) for x in other]
    best = [hash(x) for x in best]
    return ndcg(best, other)
        
    

def get_rankings(df, gt_col, est_col, query_col='query', id_col='pid'):
    """
    Does grouping, aggregation, and sorting of df to get rankings.
    df - a pyspark dataframe
    gt_col - string, the name of the column containing the ground truth relevance for query-pid pair
    est_col - string, the name of the column containing the estimated relevance for query-pid pair
    query_col - string (default 'query'), the name of the column containting the query or search terms or whatever
    id_col - string (default 'pid'), the name of the column containing item id

    returns:
    a pyspark dataframe with a single row per unique query, with columns:
      "query" (string) the query,
      "best_order" (array), list of pids ordered by gt relevance,
      "estimated_order" (array), list of pids ordered by estimated relevance
    """

    sorted_sdf = ( df.groupBy(query_col)
                    .agg(
                    F.sort_array( F.collect_list( F.struct( F.col(gt_col), F.col(id_col)) ), asc = False).alias('best'),
                    F.sort_array( F.collect_list( F.struct( F.col(est_col), F.col(id_col)) ), asc = False).alias('estimated') )  
               )

    return sorted_sdf.select(query_col, 'best.{}'.format(id_col), 'estimated').withColumnRenamed(id_col, 'best_order').select(query_col, 'best_order', 'estimated.{}'.format(id_col)).withColumnRenamed(id_col, 'estimated_order')

ndcg_udf = F.udf(ndcg, T.FloatType())


def dcg_with_rel(relevances, k=None):
    """
    Computes discounted cumulative gain
    given ground-truth relevance values.
    if k is None, computes for all relevances.
    """
    relevances = np.array(relevances)
    if k is None:
        k = len(relevances)
    
    relevances = relevances[:k]
    
    i = np.arange(1, len(relevances)+1)
    
    brp = (2.0 ** relevances)
    
    terms = (brp - 1)/np.log2(i + 1)
    score = terms.sum()
    
    return score

def ndcg_at_k(relevances, k=None):
    """
    Pass in an ordered list of relevances
    and k; returns normalized ndcg score.
    """
    relevances = np.array(relevances)
    best = relevances.copy()
    best[::-1].sort()
    best_dcg = dcg_with_rel(best, k=k)
    if best_dcg == 0.0:
        return 0.0
    dcg = dcg_with_rel(relevances, k=k)
    ndcg = dcg/best_dcg
    return float(ndcg)

def ndcg_at_k_constructor(k):
    """
    Creates and returns a pyspark udf
    that computes ndcg@k for the k passed as argument.
    """

    ndcg_udf = F.udf(lambda x: ndcg_at_k(x, k=k), T.FloatType())
    return ndcg_udf

def ranking_ndcg_at_k_constructor(k):
    """
    Creates and returns a pyspark udf
    that computes ranking ndcg@k for the k passed as argument.
    """

    ndcg_udf = F.udf(lambda x, y: ranking_ndcg_at_k(x, y, k=k), T.FloatType())
    return ndcg_udf



import numpy as np

import pyspark.sql.functions as F
import pyspark.sql.types as T
from pyspark.ml.feature import IDF, Tokenizer, CountVectorizer

def isin(element, test_elements, assume_unique=False, invert=False):
    """
    Impliments the numpy function isin() for legacy versions of
    the library
    """
    element = np.asarray(element)
    return np.in1d(element, test_elements, assume_unique=assume_unique,
                invert=invert).reshape(element.shape)

udf_template = """
def bm25(query, tf, idf):
    
    mean_dl = {}
    k = {}
    b = {}
    
    idf_values, tf_indices, tf_values, query_indices = idf.values, tf.indices, tf.values, query.indices
    freq_nidx = tf_indices[-1]+1
    freq_indices = np.concatenate((tf_indices, np.array([freq_nidx])))
    term_frequencies = np.concatenate((tf_values, np.array([0.0])))
    
    #get idf vector
    idf_ = idf_values
    
    #get term frequencies
    intersect = np.intersect1d(query_indices, freq_indices)
    idx = np.where(isin(query_indices, intersect), query_indices, freq_nidx)
    freq_idx = np.searchsorted(freq_indices, idx)
    tf_ = term_frequencies[freq_idx].reshape(-1)
    
    #get doc length
    dl_ = tf_values.sum()
    
    #get scores used to compute bm25
    ntf_ = tf_ / dl_
    ntf_score = ntf_.sum()
    tf_score = tf_.sum()
    tfidf_score = np.dot(ntf_, idf_)
    
    #get bm25
    n_term = k * (1 - b + b * dl_/mean_dl)
    bm25 = np.dot(idf_, (tf_ * (k + 1)) / ((tf_) + n_term))
    
    #return all scores
    return T.Row('tf', 'ntf', 'tfidf', 'bm25')(float(tf_score), float(ntf_score), float(tfidf_score), float(bm25))
    
schema = T.StructType([
    T.StructField("tf", T.FloatType(), False),
    T.StructField("ntf", T.FloatType(), False),
    T.StructField("tfidf", T.FloatType(), False),
    T.StructField("bm25", T.FloatType(), False)])
    
self.udf = F.udf(bm25, returnType=schema)
"""

class BM25Model(object):
    """
    Computes BM25 score.
    """
    def __init__(self, k=1.2, b=.75):
        self.k = k
        self.b = b
        self.tok = Tokenizer(inputCol='__input', outputCol='__tokens')
        self.vec = CountVectorizer(inputCol='__tokens', outputCol='__counts')
        self.idf = IDF(inputCol='__counts', outputCol='__idf')
        self.train_col = None
        self.udf = None
        self.is_fit = False
        
    def fit(self, df, train_col):
        """
        Does fitting on input df.
            df: a pyspark dataframe.
            train_col (string): The name of the column containing training documents.
            
        Returns: self, a 
        """
        self.train_col = train_col
        df_ = self.tok.transform(df.withColumnRenamed(train_col, '__input'))
        mean_dl = df_.select(F.mean(F.size(F.col('__tokens')))).collect()[0][0]
        self.vec = self.vec.fit(df_)
        df_ = self.vec.transform(df_)
        self.idf = self.idf.fit(df_)
        #this will reset value of self.udf to be a working udf function.
        exec(udf_template.format(mean_dl, self.k, self.b))
        self.is_fit = True
        return self
        
    def transform(self, df, score_col, bm25_output_name='bm25', tf_output_name=None, ntf_output_name=None, tfidf_output_name=None):
        """
        Computes BM25 score, 
            along with normalized term frequency (ntf) and tfidf.
            These three additional scores come "for free" with bm25
            but are only returned optionally.
        """
        if not self.is_fit:
            raise Exception("You must fit the BM25 model with a call to .fit() first.")
        columns = df.columns
        df_ = self.tok.transform(df.withColumnRenamed(score_col, '__input'))
        df_ = self.vec.transform(df_)
        df_ = self.idf.transform(df_)
        df_ = (df_.withColumnRenamed('__counts', '__query_counts')
               .withColumnRenamed('__input', score_col)
              ).select(columns + [score_col, '__query_counts', '__idf'])
        df_ = self.tok.transform(df_.withColumnRenamed(self.train_col, '__input'))
        df_ = self.vec.transform(df_)
        df_ = df_.withColumnRenamed('__counts', '__item_counts')
        df_ = df_.withColumn('bm25', self.udf(F.col('__query_counts'), F.col('__item_counts'), F.col('__idf')))
        df_ = df_.withColumnRenamed('__input', self.train_col)
        computed_values = df_.withColumn('more', F.explode(F.array(F.col('bm25')))).select(columns + ['bm25.*'])
        
        #this is logic for naming output column(s)
        final_selection = columns
        if bm25_output_name is not None:
            computed_values = computed_values.withColumnRenamed('bm25', bm25_output_name)
            final_selection.append(bm25_output_name)
        if tf_output_name is not None:
            computed_values = computed_values.withColumnRenamed('tf', tf_output_name)
            final_selection.append(tf_output_name)
        if ntf_output_name is not None:
            computed_values = computed_values.withColumnRenamed('ntf', ntf_output_name)
            final_selection.append(ntf_output_name)
        if tfidf_output_name is not None:
            computed_values = computed_values.withColumnRenamed('tfidf', tfidf_output_name)
            final_selection.append(tfidf_output_name)
        
        return computed_values.select(final_selection)


import numpy as np
import pandas as pd
from sklearn.metrics import mean_absolute_error
from scipy.sparse import coo_matrix

class ClickModel(object):

    def __init__(
            self,
            data,
            query_col='query',
            result_col='result',
            position_col='position',
            sessions_count='n_sessions',
            events_count='n_clicks',
            history=True,
            strategy='prior',
            val_frac=None,
            verbose=True
        ):
        if val_frac is not None:
            self.val_frac = val_frac
            self.val = True
        else:
            self.val = False
        self._init_graph(data, query_col, result_col, position_col, sessions_count, events_count)
        pos_size = len(self.pos_dict)
        qr_size = len(self.qr_dict)
        self._init_parameters(pos_size, qr_size, strategy=strategy)
        self.history = history
        if self.history:
            self.pos_diff_history = []
            self.qr_diff_history = []
            self.train_loss = []
            self.val_loss = []

        self.result_col = result_col
        self.query_col = query_col
        self.position_col = position_col
        self.verbose = verbose


    def _init_graph(self, data, query_col, result_col, position_col, sessions_count, events_count):
        """
        Initialize the sparse graph for more efficient computation.
        The resulting graph is written to self.graph and is an array like:
        [
            [position index, query-result index, number of positive events, number of negative events],
            [...],
            ...
        ]

        """
        df = data.copy()
        if 'float' in str(df[result_col].dtype):
            df[result_col] = df[result_col].astype(int)
        df['__qr_pair'] = df.apply(lambda row: str(row[query_col]) + '_' + str(row[result_col]), axis=1)
        pos_list = df[position_col].unique()
        pos_list.sort()
        self.pos_dict = {x:x for x in pos_list}

        qr_list = df['__qr_pair'].unique()
        self.qr_dict = {val:i for i,val in enumerate(qr_list)}
        df['__pos_idx'] = df[position_col].apply(lambda x: self.pos_dict[x])
        df['__qr_idx'] = df['__qr_pair'].apply(lambda x: self.qr_dict[x])
        df['__n_no_clicks'] = df[sessions_count] - df[events_count]
        graph = df[['__pos_idx', '__qr_idx', events_count, '__n_no_clicks']].values

        if self.val:
            #get validation set
            n_val_nodes = max(round(self.val_frac * len(graph)), 1)
            qrs, counts = np.unique(graph[:, 1], return_counts=True)
            to_take = qrs[np.where(counts>1, True, False)]
            try:
                to_take = np.random.choice(to_take, size=(n_val_nodes,), replace=False)
            except ValueError:
                msg = 'Cannot hold out this many nodes. Please reduce val_frac.'
                raise Exception(msg)
            val_idx = []
            for qr_idx in to_take:
                candidates = np.argwhere(graph[:,1]==qr_idx).flatten()
                idx = np.random.choice(candidates)
                val_idx.append(idx)
            val_idx = np.array(val_idx)

            self.val_graph = graph[val_idx]
            self.graph = graph[~np.isin(np.arange(0, len(graph)), val_idx)]
        else:
            self.graph=graph

    def _init_parameters(self, pos_size, qr_size, strategy):
        if strategy == 'random':
            self.pos = np.random.random(size=(pos_size,))
            self.qr = np.random.random(size=(qr_size,))
        elif isinstance(strategy, float):
            self.pos = np.array([strategy for _ in range(pos_size)])
            self.qr = np.array([.1 for _ in range(qr_size)])
        elif strategy == 'prior':
            pos_idx = self.graph[:, 0]
            qr_idx = self.graph[:, 1]
            n_clicks = self.graph[:, 2]
            nums = np.array(coo_matrix((n_clicks, (qr_idx, pos_idx))).sum(axis=0)).reshape(-1)
            n_no_clicks = self.graph[:, 3]
            denoms = np.array(coo_matrix((n_no_clicks, (qr_idx, pos_idx))).sum(axis=0)).reshape(-1)
            denoms = denoms + nums
            estimate = nums/denoms
            self.pos = estimate * (.99/estimate.max())

            ctrs = n_clicks / (n_clicks+n_no_clicks)
            P_inspect = self.pos[pos_idx]
            P_click = ctrs/P_inspect
            #constrain value between .01 and .99
            P_click = np.where(P_click > .99, .99, P_click)
            P_click = np.where(P_click < .01, .01, P_click)
            weights = n_clicks + n_no_clicks
            nums = np.array(coo_matrix((P_click * weights, (qr_idx, pos_idx))).sum(axis=1)).reshape(-1)
            denoms = np.array(coo_matrix((weights, (qr_idx, pos_idx))).sum(axis=1)).reshape(-1)

            self.qr = nums/denoms
        else:
            self.pos = np.random.random(size=(pos_size,))
            self.qr = np.random.random(size=(qr_size,))

    def _compute_update(self, kind=None):
        #these values are shared in both updates
        qr_idx = self.graph[:, 1]
        alpha = self.qr[qr_idx]
        pos_idx = self.graph[:, 0]
        gamma = self.pos[pos_idx]
        ctr = self.graph[:, 2]
        no_ctr = self.graph[:, 3]
        #these terms appears in both updates
        shared_term = (1 - gamma * alpha)
        denom = ctr + no_ctr
        denom_mat = coo_matrix((denom, (qr_idx, pos_idx)))

        if kind == 'qr' or kind is None:
            #here we compute update to alpha params
            qr_update = (ctr + (no_ctr * (((1 - gamma) * alpha) / shared_term)))
            qr_mat = coo_matrix((qr_update, (qr_idx, pos_idx)))
            qr_vec = qr_mat.sum(axis=1)

            denom_mat = coo_matrix((denom, (qr_idx, pos_idx)))
            qr_denom_vec = denom_mat.sum(axis=1)

            #this vector is the new set of alpha params
            new_qr = np.array(qr_vec / qr_denom_vec).reshape(-1)

            if kind == 'qr':
                return new_qr

        if kind == 'pos' or kind is None:

            #here we compute update to gamma params
            pos_update = (ctr + (no_ctr * (((1 - alpha) * gamma) / shared_term)))
            pos_mat = coo_matrix((pos_update, (qr_idx, pos_idx)))
            pos_vec = pos_mat.sum(axis=0)

            pos_denom_vec = denom_mat.sum(axis=0)

            #this vector is the new set of alpha params
            new_pos = np.array((pos_vec / pos_denom_vec)).reshape(-1)

            if kind == 'pos':
                return new_pos

        #return both if none specified
        return new_pos, new_qr

    def compute_diff(self, v0, v1):
        return np.linalg.norm(v0 - v1)

    def update_pos(self):
        new_pos = self._compute_update(kind='pos')
        self.pos = new_pos
        diff_pos = self.compute_diff(self.pos, new_pos)
        if self.history:
            self.pos_diff_history.append(diff_pos)

    def update_qr(self):
        new_qr = self._compute_update(kind='qr')
        self.qr = new_qr
        diff_qr = self.compute_diff(self.qr, new_qr)
        if self.history:
            self.qr_diff_history.append(diff_qr)

    def update_both(self):
        new_pos, new_qr = self._compute_update()
        diff_pos = self.compute_diff(self.pos, new_pos)
        diff_qr = self.compute_diff(self.qr, new_qr)
        if self.history:
            self.pos_diff_history.append(diff_pos)
            self.qr_diff_history.append(diff_qr)
        self.pos = new_pos
        self.qr = new_qr

    def validate(self, val_set=False):
        if val_set:
            graph = self.val_graph
        else:
            graph = self.graph
        total = graph[:, 2] + graph[:, 3]
        positive_outcomes = graph[:, 2]
        predicted_outcomes = (self.pos[graph[:, 0]] * self.qr[graph[:, 1]]) * total
        return mean_absolute_error(positive_outcomes, predicted_outcomes)

    def do_update(self, alternate=True):
        if alternate:
            self.update_pos()
            self.update_qr()
        else:
            self.update_both()

    def decode_qr_pairs(self, items):
        split = [x.split('_') for x in items]
        terms = ['_'.join(x[:-1]) for x in split]
        results = [x[-1] for x in split]
        return results, terms

    def fit(self, n_iterations=1000, alternate=True, stopping=None, min_iterations=20):
        if stopping is not None:
            prev = 1e99
        for i in range(n_iterations):
            self.do_update(alternate)
            val_loss = self.validate(val_set=self.val)
            if self.val:
                train_loss = self.validate(val_set=False)
            else:
                train_loss = val_loss
            if i%10 == 0 and self.verbose:
                print('validation loss at iteration {}: '.format(i), round(val_loss, 4))
            if self.history:
                self.train_loss.append(train_loss)
                self.val_loss.append(val_loss)
            if stopping is not None:
                diff = prev - val_loss
                if i > min_iterations and diff < stopping:
                    if self.verbose:
                        print('stopping early with ', str(i), ' iterations  and a val loss of ', val_loss)
                    break
                prev = val_loss

        qr_map = {self.qr_dict[key]:key for key in self.qr_dict}
        qr_keys = [qr_map[i] for i in range(len(self.qr))]
        pids, results = self.decode_qr_pairs(qr_keys)
        qr_values = self.qr
        qr_results = pd.DataFrame(columns=[self.result_col, self.query_col, 'attractiveness'])
        qr_results[self.result_col] = pids
        qr_results[self.query_col] = results
        qr_results['attractiveness'] = qr_values

        pos_results = pd.DataFrame(columns=[self.position_col, 'p_examine'])
        pos_results[self.position_col] = list(range(len(self.pos)))
        pos_results['p_examine'] = self.pos

        if self.history:
            history = {
                'val_loss':self.val_loss,
                'train_loss':self.train_loss,
                'delta_pos':self.pos_diff_history,
                'delta_qr':self.qr_diff_history
            }

        else:
            history = {}

        return qr_results, pos_results, history



from setuptools import find_packages, setup
import sys

requirements = [
    'numpy>=1.13.0',
    'scipy',
    'pandas',
    'pyspark',
    'scikit-learn'
]

#public version 0.0.7
version = '0.0.7'

setup(
    name='search-tools',
    version=version,
    author="Michael Klear",
    author_email='fixityourself@fu.com',
    url='https://github.com/AlliedToasters/search-tools/archive/v0.0.7.tar.gz',
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements
)



import unittest
import numpy as np
import pandas as pd

from search_tools.metrics import ndcg_at_k, dcg_with_rel, ranking_ndcg_at_k
from search_tools.metrics import ndcg_at_k_constructor, ranking_ndcg_at_k_constructor
from search_tools.models import ClickModel

np.random.seed(42)

def generate_data(n_events=100):
    pos_distr = np.random.uniform(low=0.1, high=0.99, size=(4,))
    queries = np.array(['do a search', 'i like to search', 'things for me', 'shopping item'])
    results = np.array(['doc0', 'doc1', 'doc2', 'doc3', 'doc4', 'doc5', 'doc6', 'doc7'])
    affinities = pd.DataFrame(columns=queries, index=results, data=np.random.uniform(low=0.1, high=0.99, size=(len(results), len(queries))))
    cols = ['_query', '_result', '_position', '_click']
    outcomes = {
        'do a search':[
            ['doc0', 'doc2', 'doc4', 'doc5'],
            ['doc2', 'doc0', 'doc5', 'doc6'],
            ['doc5', 'doc6', 'doc0', 'doc4']
        ],
        'i like to search':[
            ['doc2', 'doc7', 'doc6', 'doc1'],
            ['doc2', 'doc1', 'doc0', 'doc7'],
            ['doc1', 'doc6', 'doc7', 'doc2']
        ],
        'things for me':[
            ['doc5', 'doc4', 'doc3', 'doc2'],
            ['doc3', 'doc2', 'doc5', 'doc6'],
            ['doc2', 'doc3', 'doc4', 'doc5']
        ],
        'shopping item':[
            ['doc0', 'doc1', 'doc3', 'doc2'],
            ['doc5', 'doc2', 'doc3', 'doc0'],
            ['doc5', 'doc0', 'doc1', 'doc2']
        ]
    }

    dataset = pd.DataFrame(columns=cols)

    for i in range(n_events):
        search_term = np.random.choice(queries)
        returned_results = np.random.randint(0, 3)
        results = outcomes[search_term][returned_results]
        rows = pd.DataFrame(columns=cols)
        rows['_result'] = results
        rows['_query'] = search_term
        rows['_position'] = np.arange(4)
        ps = []
        for i, row in rows.iterrows():
            p_pos = pos_distr[i]
            p_qr = affinities.loc[row['_result'], row['_query']]
            ps.append(p_qr * p_pos)
        outcome = np.random.random(size=(4,))
        rows['_click'] = (ps > outcome).astype(int)
        dataset = pd.concat([dataset, rows], axis=0)

    clicks = dataset.groupby(['_query', '_result', '_position']).apply(lambda x: x['_click'].sum())
    count = dataset.groupby(['_query', '_result', '_position']).count()

    df = pd.concat([clicks, count], axis=1)
    df.columns = ['total_clicks', 'total_sessions']
    return df.reset_index(), pos_distr, affinities

class ModelsTest(unittest.TestCase):

    def setUp(self):
        self.df, _, _ = generate_data()

    def test_run_click_model(self):
        data = self.df
        model = ClickModel(
            data,
            query_col='_query',
            result_col='_result',
            position_col='_position',
            sessions_count='total_sessions',
            events_count='total_clicks',
            history=True,
            strategy='prior',
            verbose=False
        )
        qr, pos, history = model.fit(n_iterations=4, alternate=True)
        assert isinstance(qr, pd.DataFrame)
        assert isinstance(pos, pd.DataFrame)

    def test_stopping_click_model(self):
        data = self.df
        model = ClickModel(
            data,
            query_col='_query',
            result_col='_result',
            position_col='_position',
            sessions_count='total_sessions',
            events_count='total_clicks',
            history=True,
            strategy='prior',
            val_frac=.1,
            verbose=False
        )
        qr, pos, history = model.fit(n_iterations=200000, alternate=True, stopping=1e-5)
        assert isinstance(qr, pd.DataFrame)
        assert isinstance(pos, pd.DataFrame)

    def test_val_exception(self):
        data = self.df
        try:
            model = ClickModel(
                data,
                query_col='_query',
                result_col='_result',
                position_col='_position',
                sessions_count='total_sessions',
                events_count='total_clicks',
                history=True,
                strategy='prior',
                val_frac=.1,
                verbose=False
            )
            raise AssertionError('Expected failure due to large frac setting.')
        except AssertionError:
            pass

    def test_no_val(self):
        data = self.df
        model = ClickModel(
            data,
            query_col='_query',
            result_col='_result',
            position_col='_position',
            sessions_count='total_sessions',
            events_count='total_clicks',
            history=True,
            strategy='prior',
            verbose=False
        )
        model.fit(3)




class MetricsTest(unittest.TestCase):

    def test_dcg_with_rel(self):
        rels = [.5, .3, .6, .8]
        result = dcg_with_rel(rels)
        assert np.abs((result - 1.137)) < 0.01


    def test_ndcg_at_k(self):
        rels = [.5, .3, .6, .8]
        result = ndcg_at_k(rels)
        best = 1.373
        expected = 1.137/best
        assert np.abs((result - expected)) < 0.01

    def test_ndcg_at_k_with_k(self):
        rels = [.5, .3, .6, .8]
        result = ndcg_at_k(rels, k=4)
        best = 1.373
        expected = 1.137/best
        assert np.abs((result - expected)) < 0.01

    def test_big_k(self):
        rels = [.5, .3, .6, .8]
        result = ndcg_at_k(rels, k=10)
        best = 1.373
        expected = 1.137/best
        assert np.abs((result - expected)) < 0.01

    def test_little_k(self):
        rels = [.5, .3, .6, .8]
        result = ndcg_at_k(rels, k=2)
        best = 1.373
        expected = 1.137/best
        assert np.abs((result - expected)) > 0.01

    def test_ndcgatk_with_zeros(self):
        rels = [0, 0, 0, 0]
        result = ndcg_at_k(rels, k=None)
        assert result == 0

    def test_ordered_ndcgatk(self):
        best = [123, 332, 101, 100, 777, 9, 456]
        best = [str(x) for x in best]
        other = [101, 456, 332, 100, 999]
        other = [str(x) for x in other]

        return ranking_ndcg_at_k(best, other, k=None)

    def test_constructors(self):
        func = ndcg_at_k_constructor(k=5)
        func = ranking_ndcg_at_k_constructor(k=5)

if __name__ in "__main__":
    unittest.main()
